"""Basic constants/names used throughout the lib."""
import numpy as np

#: Byte information of each sensor_type in one sample.
#:
#: Format: Overall number of bytes, number of channels, datatype
SENSOR_SAMPLE_LENGTH = {
    "acc": (6, 3, np.int16),
    "gyro": (6, 3, np.int16),
    "mag": (6, 3, np.int16),
    "baro": (2, 1, np.int16),
    "analog": (3, 3, np.uint8),
    "ecg": (4, 1, np.int32),
    "ppg": (4, 1, np.int32),
    "temperature": (2, 1, np.int16),
    "counter": (4, 1, float),
}

#: Default legends for all sensors
SENSOR_LEGENDS = {
    "acc": tuple("acc_" + x for x in "xyz"),
    "gyro": tuple("gyr_" + x for x in "xyz"),
    "baro": tuple(["baro"]),
    "analog": tuple("analog_" + str(x) for x in range(3)),
    "ecg": tuple(["ecg"]),
    "ppg": tuple(["ppg"]),
    "temperature": tuple(["temp"]),
}

#: Default units for all sensors
SENSOR_UNITS = {"acc": "g", "gyro": "dps", "baro": "mbar", "temperature": "C"}

#: Available sensor_type positions
SENSOR_POS = {0: "undefined", 1: "left foot", 2: "right foot", 3: "hip", 4: "left wrist", 5: "right wrist", 6: "chest"}



# -*- coding: utf-8 -*-
"""Fundamental Datastream class, which holds any type of sensor_type data and handles basic interactions with it."""

import copy
from typing import Optional, Iterable, List, TypeVar, TYPE_CHECKING

import numpy as np
from scipy.signal import resample

from nilspodlib.consts import SENSOR_LEGENDS, SENSOR_UNITS
from nilspodlib.utils import inplace_or_copy

T = TypeVar("T")

if TYPE_CHECKING:
    import pandas as pd  # noqa: F401


class Datastream:
    """Object representing a single set of data of one sensor_type.

    Usually it is not required to directly interact with the datastream object (besides accessing the data attribute).
    Most important functionality can/should be used via a dataset or session object to manage multiple datasets at once.

    Attributes
    ----------
    data
        The actual data of the sensor_type as `np.array`.
        Can have multiple dimensions depending on the sensor_type.
    sensor
        The name of the sensor_type
    is_calibrated
        If the sensor_type is in a raw format or the expected final output units
    sampling_rate_hz
        The sampling rate of the datastream

    """

    data: np.ndarray
    is_calibrated: bool = False
    sampling_rate_hz: float
    sensor_type: Optional[str]
    _unit: str
    _columns: Optional[List]

    def __init__(
        self,
        data: np.ndarray,
        sampling_rate: float = 1.0,
        columns: Optional[Iterable] = None,
        unit: Optional[str] = None,
        sensor_type: Optional[str] = None,
    ):
        """Get new datastream instance.

        Parameters
        ----------
        data :
            The actual data to be stored in the datastream.
            Should be 2D (even if only 1D data is stored).
            First dimension should be the time axis and second dimension the different data vector entries
            (e.g. acc_x, acc_y, acc_z).
        sampling_rate :
            The sampling rate of the datastream.
            Is used for all calculations that require sampling info.
        columns :
            Optional list of names for the data vector entries.
            Only used to make it easier to understand the content of a datastream.
        unit :
            The unit of the data stored in the datastream.
            This is only used, if `self.is_calibrated` is set to `True`.
        sensor_type :
            Type of sensor_type the data is produced from.
            This allows to automatically get default values for columns and units from :py:mod:`nilspodlib.consts`.

        """
        self.data = data
        self.sampling_rate_hz = float(sampling_rate)
        self.sensor_type = sensor_type
        self._columns = list(columns) if columns else columns
        self._unit = unit

    def __repr__(self):
        """Provide a meaningful str-representation of a Datastream."""
        return "Datastream(sensor_type={}, sampling_rate_hz={}, is_calibrated={}, data={}".format(
            self.sensor_type, self.sampling_rate_hz, self.is_calibrated, self.data
        )

    @property
    def unit(self):
        """Get the unit of the data contained in the datastream.

        This will return either `a.u.` if the datastream is not yet calibrated, the value of `self._unit` if set or
        the default unit from :py:data:`nilspodlib.consts.SENSOR_UNITS`
        """
        if self.is_calibrated is True:
            if self._unit:
                return self._unit
            if self.sensor_type and SENSOR_UNITS.get(self.sensor_type, None):
                return SENSOR_UNITS[self.sensor_type]
        return "a.u."

    @property
    def columns(self):
        """Get the column headers for the data contained in the datastream.

        This will return `self._columns` if set on init or will try to get the default columns from
        :py:data:`nilspodlib.consts.SENSOR_LEGENDS`.
        If none of the above is applicable the columns wil be numbered starting with 0.
        """
        if self._columns:
            return self._columns
        if self.sensor_type:
            if SENSOR_LEGENDS.get(self.sensor_type, None):
                return list(SENSOR_LEGENDS[self.sensor_type])
        return list(range(self.data.shape[-1]))

    def __len__(self):
        """Length of the datastream in samples."""
        return len(self.data)

    def norm(self) -> np.ndarray:
        """Calculate the norm of the data along the last axis.

        This will provide the vector norm for each time point.
        """
        return np.linalg.norm(self.data, axis=-1)

    def normalize(self) -> "Datastream":
        """Get the normalized data.

        Normalization is performed by dividing the data by its maximum absolute value.
        This will always return a copy of the datastream.
        """
        ds = copy.deepcopy(self)
        ds.data /= np.abs(ds.data).max(axis=0)
        return ds

    def cut(
        self: T,
        start: Optional[int] = None,
        stop: Optional[int] = None,
        step: Optional[int] = None,
        inplace: bool = False,
    ) -> T:
        """Cut the datastream.

        This is equivalent to applying the following slicing to the data of the datastream: `array[start:stop:step]`

        Parameters
        ----------
        start :
            Start index
        stop :
            Stop index
        step :
            Step size of the cut
        inplace :
            If True this methods modifies the current datastream object. If False, a copy of the datastream is returned.

        """
        s = inplace_or_copy(self, inplace)
        sl = slice(start, stop, step)
        s.data = s.data[sl]
        return s

    def downsample(self: T, factor: int, inplace: bool = False) -> T:
        """Downsample the datastreams by a factor.

        Parameters
        ----------
        factor :
            Factor by which the datastream should be downsampled.
        inplace :
            If True this methods modifies the current datastream object. If False, a copy of the datastream is returned.

        """
        s = inplace_or_copy(self, inplace)
        s.data = resample(s.data, len(s.data) // factor, axis=0)
        s.sampling_rate_hz /= factor
        return s

    def data_as_df(self, index_as_time: bool = True, include_units=False) -> "pd.DataFrame":
        """Export the datastream as pandas DataFrame using `self.columns` as colummns.

        Parameters
        ----------
        index_as_time :
            If True the index will be divided by the sampling rate to represent time since start of the measurement
        include_units :
            If True the column names will have the unit of the datastream concatenated with an (Default value = False)

        Returns
        -------
        DataFrame

        """
        import pandas as pd  # noqa: F811

        columns = self.columns
        if include_units is True:
            columns = ["{}_{}".format(c, self.unit) for c in columns]
        df = pd.DataFrame(self.data, columns=columns)
        if index_as_time:
            df.index /= self.sampling_rate_hz
            df.index.name = "t"
        return df



# -*- coding: utf-8 -*-
"""A python package to parse logged NilsPod Binary files.

@author: Arne Küderle, Nils Roth
"""
from .dataset import Dataset  # noqa: F401
from .session import Session, SyncedSession  # noqa: F401

__all__ = ["Dataset", "Session", "SyncedSession"]



"""A set of utilities to save and find calibrations for NilsPod sensors."""

import datetime
import json
import re
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Union

import numpy as np

from nilspodlib.exceptions import CalibrationWarning
from nilspodlib.utils import path_t

if TYPE_CHECKING:
    from imucal import CalibrationInfo  # noqa: F401


def save_calibration(
    calibration: "CalibrationInfo", sensor_id: str, cal_time: datetime.datetime, folder: path_t
) -> Path:
    """Save a calibration info object in the correct format and file name for NilsPods.

    The files will be saved in the format: `folder / {sensor_id}_%Y-%m-%d_%H-%M.json`

    The naming schema and format is of course just a suggestion, and any structure can be used as long as it can be
    converted back into a CalibrationInfo object.
    However, following the naming convention will allow to use other calibration utils to search for suitable
    calibration files.

    .. note:: If the folder does not exist it will be created.

    Parameters
    ----------
    calibration :
        The CalibrationInfo object ot be saved
    sensor_id :
        The for 4 letter/digit identifier of a sensor_type, as obtained from
        :py:meth:`nilspodlib.header.Header.sensor_id`
    cal_time :
        The date and time (min precision) when the calibration was performed. It is preferable to pass this
        value in UTC timezone, as this is in line with the time handling in the rest of the library.
    folder :
        Basepath of the folder, where the file will be stored.

    Returns
    -------
    output_file_name
        The name under which the calibration file was saved

    """
    if not re.fullmatch(r"\w{4}", sensor_id):
        raise ValueError(
            "The sensor_id is expected to be a 4 symbols string only containing numbers or letters, not {}".format(
                sensor_id
            )
        )
    Path(folder).mkdir(parents=True, exist_ok=True)
    f_name = Path(folder) / "{}_{}.json".format(sensor_id.lower(), cal_time.strftime("%Y-%m-%d_%H-%M"))
    calibration.to_json_file(f_name)
    return f_name


def find_calibrations_for_sensor(
    sensor_id: str,
    folder: Optional[path_t] = None,
    recursive: bool = True,
    filter_cal_type: Optional[str] = None,
    ignore_file_not_found: Optional[bool] = False,
) -> List[Path]:
    """Find possible calibration files based on the filename.

    As this only checks the filenames, this might return false positives depending on your folder structure and naming.

    Parameters
    ----------
    sensor_id :
        The for 4 letter/digit identifier of a sensor_type, as obtained from
        :py:meth:`nilspodlib.header.Header.sensor_id`
    folder :
        Basepath of the folder to search. If None, tries to find a default calibration
    recursive :
        If the folder should be searched recursive or not.
    filter_cal_type :
        Whether only files obtain with a certain calibration type should be found.
        This will look for the `CalType` inside the json file and hence cause performance problems.
        If None, all found files will be returned.
        For possible values, see the `imucal` library.
    ignore_file_not_found :
        If True this function will not raise an error, but rather return an empty list, if no
        calibration files were found for the specific sensor_type.

    Returns
    -------
        list_of_cals
            List of paths pointing to available calibration objects.

    """
    if not folder:
        try:
            from NilsPodRefCal import CAL_PATH  # noqa: import-outside-toplevel
        except ImportError:
            raise ImportError(
                "The module NilsPodRefCal is not installed. If you want support for default calibrations,"
                " please install it from"
                " https://mad-srv.informatik.uni-erlangen.de/MadLab/portabilestools/nilspodrefcal"
            )
        folder = CAL_PATH

    method = "glob"
    if recursive is True:
        method = "rglob"

    r = sensor_id.lower() + r"_\d{4}-\d{2}-\d{2}_\d{2}-\d{2}"

    matches = [f for f in getattr(Path(folder), method)("{}_*.json".format(sensor_id)) if re.fullmatch(r, f.stem)]

    if filter_cal_type:
        matches = [f for f in matches if json.load(f.open())["cal_type"].lower() == filter_cal_type.lower()]

    if not matches and ignore_file_not_found is not True:
        raise ValueError("No Calibration for the sensor_type with the id {} could be found".format(sensor_id))
    return matches


def find_closest_calibration_to_date(
    sensor_id: str,
    cal_time: datetime.datetime,
    folder: Optional[path_t] = None,
    recursive: bool = True,
    filter_cal_type: Optional[str] = None,
    before_after: Optional[str] = None,
    warn_thres: datetime.timedelta = datetime.timedelta(days=30),  # noqa E252
    ignore_file_not_found: Optional[bool] = False,
) -> Optional[Path]:
    """Find the calibration file for a sensor_type, that is closes to a given date.

    As this only checks the filenames, this might return a false positive depending on your folder structure and naming.

    Parameters
    ----------
    sensor_id :
        The for 4 letter/digit identifier of a sensor_type, as obtained from
        :py:meth:`nilspodlib.header.Header.sensor_id`
    cal_time :
        time and date to look for
    folder :
        Basepath of the folder to search. If None, tries to find a default calibration
    recursive :
        If the folder should be searched recursive or not.
    filter_cal_type :
        Whether only files obtain with a certain calibration type should be found.
        This will look for the `CalType` inside the json file and hence cause performance problems.
        If None, all found files will be returned.
        For possible values, see the `imucal` library.
    before_after :
        Can either be 'before' or 'after', if the search should be limited to calibrations that were
        either before or after the specified date. If None the closest value ignoring if it was before or after the
        measurement.
    warn_thres :
        If the distance to the closest calibration is larger than this threshold, a warning is emitted
    ignore_file_not_found :
        If True this function will not raise an error, but rather return `None`, if no
        calibration files were found for the specific sensor_type.

    Notes
    -----
    If there are multiple calibrations that have the same date/hour/minute distance form the measurement,
    the calibration before the measurement will be chosen. This can be overwritten using the `before_after` para.

    See Also
    --------
    nilspodlib.calibration_utils.find_calibrations_for_sensor

    Returns
    -------
    cal_file_path or None
        The path to a suitable calibration file, or `None`, if no suitable file could be found.

    """
    if before_after not in ("before", "after", None):
        raise ValueError('Invalid value for `before_after`. Only "before", "after" or None are allowed')

    potential_list = find_calibrations_for_sensor(
        sensor_id=sensor_id,
        folder=folder,
        recursive=recursive,
        filter_cal_type=filter_cal_type,
        ignore_file_not_found=ignore_file_not_found,
    )
    if not potential_list:
        if ignore_file_not_found is True:
            return None
        raise ValueError("No Calibration for the sensor_type with the id {} could be found".format(sensor_id))

    dates = [datetime.datetime.strptime("_".join(d.stem.split("_")[1:]), "%Y-%m-%d_%H-%M") for d in potential_list]

    dates = np.array(dates, dtype="datetime64[s]")
    potential_list, _ = zip(*sorted(zip(potential_list, dates), key=lambda x: x[1]))
    dates.sort()

    diffs = (dates - np.datetime64(cal_time, "s")).astype(float)

    if before_after == "after":
        diffs[diffs < 0] = np.nan
    elif before_after == "before":
        diffs[diffs > 0] = np.nan

    if np.all(diffs) == np.nan:
        raise ValueError("No calibrations {} {} were found.".format(before_after, cal_time))

    min_dist = float(np.nanmin(np.abs(diffs)))
    if warn_thres < datetime.timedelta(seconds=min_dist):
        warnings.warn(
            "For the sensor_type {} no calibration could be located that was in {} of the {}."
            "The closest calibration is {} away.".format(
                sensor_id, warn_thres, cal_time, datetime.timedelta(seconds=min_dist)
            ),
            CalibrationWarning,
        )

    return potential_list[int(np.nanargmin(np.abs(diffs)))]


def load_and_check_cal_info(calibration: Union["CalibrationInfo", path_t]) -> "CalibrationInfo":
    """Load a calibration from path or check if the provided object is already a valid calibration."""
    from imucal import CalibrationInfo  # noqa: import-outside-toplevel

    if isinstance(calibration, (Path, str)):
        calibration = CalibrationInfo.from_json_file(calibration)
    if not isinstance(calibration, CalibrationInfo):
        raise ValueError("No valid CalibrationInfo object provided")
    return calibration



# -*- coding: utf-8 -*-
"""Dataset represents a measurement session of a single sensor_type."""
import datetime
import warnings
from distutils.version import StrictVersion
from pathlib import Path
from typing import Union, Iterable, Optional, Tuple, Dict, TypeVar, Type, Sequence, TYPE_CHECKING, List

import numpy as np
import pandas as pd
from scipy.signal import resample

from nilspodlib.calibration_utils import (
    find_closest_calibration_to_date,
    find_calibrations_for_sensor,
    load_and_check_cal_info,
)
from nilspodlib.consts import SENSOR_SAMPLE_LENGTH
from nilspodlib.datastream import Datastream
from nilspodlib.exceptions import (
    InvalidInputFileError,
    RepeatedCalibrationError,
    datastream_does_not_exist_warning,
    SynchronisationWarning,
    LegacyWarning,
)
from nilspodlib.header import Header
from nilspodlib.legacy import legacy_support_check, find_conversion_function
from nilspodlib.utils import (
    path_t,
    read_binary_uint8,
    convert_little_endian,
    inplace_or_copy,
    get_header_and_data_bytes,
    get_strict_version_from_header_bytes,
)

if TYPE_CHECKING:
    from imucal import CalibrationInfo  # noqa: F401

T = TypeVar("T")


class Dataset:  # noqa: too-many-public-methods
    """Class representing a logged session of a single NilsPod.

    .. warning:: Some operations on the dataset should not be performed after each other, as they can lead to unexpected
                 results.
                 The respective methods have specific warnings in their docstring.

    Each instance has 3 important (groups of attributes):

    - self.info: A instance of `nilspodlib.header.Header` containing all the meta info about the measurement.
    - self.counter: The continuous counter created by the sensor.
      It is in particular important to synchronise multiple datasets that were recorded at the same time
      (see `nilspodlib.session.SyncedSession`).
    - datastream: The actual sensor_type data accessed directly by the name of the sensor_type
      (e.g. acc, gyro, baro, ...).
      Each sensor_type data is wrapped in a `NilPodLib.datastream.Datastream` object.

    Attributes
    ----------
    path :
        Path pointing to the recording file (if dataset was loaded from a file)
    info :
        Metadata of the recording
    size :
        The number of samples in the dataset.
    counter :
        The continuous counter of the sensor.
    time_counter :
        Counter in seconds since first sample.
    utc_counter :
        Counter as utc timestamps.
    utc_datetime_counter :
        Counter as np.datetime64 in UTC timezone.
    active_sensor :
        The enabled sensors in the dataset.
    datastreams :
        Iterator over all available datastreams/sensors
    acc :
        Optional accelerometer datastream.
    gyro :
        Optional gyroscope datastream.
    mag :
        Optional magnetometer datastream.
    baro :
        Optional barometer datastream.
    analog :
        Optional analog datastream.
        Its content will depend on the exact recording and sensor used.
    ecg :
        Optional ECG datastream.
    ppg :
        Optional PPG datastream.
    temperature :
        Optional temperature reading datastream.

    """

    path: path_t
    acc: Optional["Datastream"] = None
    gyro: Optional["Datastream"] = None
    mag: Optional["Datastream"] = None
    baro: Optional["Datastream"] = None
    analog: Optional["Datastream"] = None
    ecg: Optional["Datastream"] = None
    ppg: Optional["Datastream"] = None
    temperature: Optional["Datastream"] = None
    counter: np.ndarray
    info: Header

    @property
    def size(self) -> int:
        """Get the number of samples in the Dataset."""
        return len(self.counter)

    @property
    def active_sensors(self) -> Tuple[str]:
        """Get the enabled sensors in the dataset."""
        return tuple(self.info.enabled_sensors)

    @property
    def datastreams(self) -> Iterable[Datastream]:
        """Iterate through all available datastreams."""
        for i in self.active_sensors:
            yield i, getattr(self, i)

    @property
    def utc_counter(self) -> np.ndarray:
        """Counter as utc timestamps."""
        return self.info.utc_datetime_start_day_midnight.timestamp() + self.counter / self.info.sampling_rate_hz

    @property
    def utc_datetime_counter(self) -> np.ndarray:
        """Counter as np.datetime64 in UTC timezone."""
        return (self.utc_counter * 1e6).astype("datetime64[us]")

    @property
    def time_counter(self) -> np.ndarray:
        """Counter in seconds since first sample."""
        return (self.counter - self.counter[0]) / self.info.sampling_rate_hz

    def __init__(self, sensor_data: Dict[str, np.ndarray], counter: np.ndarray, info: Header):
        """Get new Dataset instance.

        .. note::
            Usually you shouldn't use this init directly.
            Use the provided `from_bin_file` constructor to handle loading recorded NilsPod Sessions.

        Parameters
        ----------
        sensor_data :
            Dictionary with name of sensor_type and sensor_type data as np.array
            The data needs to be 2D with time/counter as first dimension
        counter :
            The counter created by the sensor_type. Should have the same length as all datasets
        info :
            Header instance containing all Metainfo about the measurement.

        """
        self.counter = counter
        self.info = info
        for k, v in sensor_data.items():
            v = Datastream(v, self.info.sampling_rate_hz, sensor_type=k)
            setattr(self, k, v)

    @classmethod
    def from_bin_file(cls: Type[T], path: path_t, legacy_support: str = "error") -> T:
        """Create a new Dataset from a valid .bin file.

        Parameters
        ----------
        path :
            Path to the file
        legacy_support :
            This indicates how to deal with old firmware versions.
            If `error`: An error is raised, if an unsupported version is detected.
            If `warn`: A warning is raised, but the file is parsed without modification
            If `resolve`: A legacy conversion is performed to load old files. If no suitable conversion is found,
            an error is raised. See the `legacy` package and the README to learn more about available
            conversions.

        Raises
        ------
        VersionError
            If unsupported FirmwareVersion is detected and `legacy_error` is True

        """
        path = Path(path)
        if path.suffix != ".bin":
            ValueError('Invalid file type! Only ".bin" files are supported not {}'.format(path))

        sensor_data, counter, info = parse_binary(path, legacy_support=legacy_support)
        s = cls(sensor_data, counter, info)

        s.path = path
        return s

    @classmethod
    def from_csv_file(cls, path: path_t):
        """Create a new Dataset from a valid .csv file.

        Parameters
        ----------
        path :
            Path to the file
        path :
            Path to the file

        Notes
        -----
        This is planned but not yet supported

        """
        raise NotImplementedError("CSV importer coming soon")

    def calibrate_imu(self: T, calibration: Union["CalibrationInfo", path_t], inplace: bool = False) -> T:
        """Apply a calibration to the Acc and Gyro datastreams.

        The final units of the datastreams will depend on the used calibration values, but must likely they will be "g"
        for the Acc and "dps" (degrees per second) for the Gyro.

        Parameters
        ----------
        calibration :
            calibration object or path to .json file, that can be used to create one.
        inplace :
            If True this methods modifies the current dataset object. If False, a copy of the dataset and all
            datastream objects is created
            Notes:
        inplace :
            If True this methods modifies the current dataset object. If False, a copy of the dataset and all
            datastream objects is created
            Notes:
            This just combines `calibrate_acc` and `calibrate_gyro`.

        """
        s = inplace_or_copy(self, inplace)
        check = [self._check_calibration(s.acc, "acc"), self._check_calibration(s.gyro, "gyro")]
        if all(check):
            calibration = load_and_check_cal_info(calibration)
            acc, gyro = calibration.calibrate(s.acc.data, s.gyro.data)
            s.acc.data = acc
            s.gyro.data = gyro
            s.acc.is_calibrated = True
            s.acc._unit = calibration.ACC_UNIT
            s.gyro.is_calibrated = True
            s.gyro._unit = calibration.GYRO_UNIT
        return s

    def calibrate_acc(self: T, calibration: Union["CalibrationInfo", path_t], inplace: bool = False) -> T:
        """Apply a calibration to the Acc datastream.

        The final units of the datastream will depend on the used calibration values, but must likely they will be "g"
        for Acc.

        Parameters
        ----------
        calibration :
            calibration object or path to .json file, that can be used to create one.
        inplace :
            If True this methods modifies the current dataset object. If False, a copy of the dataset and all
            datastream objects is created

        """
        s = inplace_or_copy(self, inplace)
        if self._check_calibration(s.acc, "acc") is True:
            calibration = load_and_check_cal_info(calibration)
            acc = calibration.calibrate_acc(s.acc.data)
            s.acc.data = acc
            s.acc.is_calibrated = True
            s.acc._unit = calibration.ACC_UNIT
        return s

    def calibrate_gyro(self: T, calibration: Union["CalibrationInfo", path_t], inplace: bool = False) -> T:
        """Apply a calibration to the Gyro datastream.

        The final units of the datastreams will depend on the used calibration values, but must likely they will be
        "dps" (degrees per second) for the Gyro.

        Parameters
        ----------
        calibration :
            calibration object or path to .json file, that can be used to create one.
        inplace :
            If True this methods modifies the current dataset object. If False, a copy of the dataset and all
            datastream objects is created

        """
        s = inplace_or_copy(self, inplace)
        if self._check_calibration(s.gyro, "gyro") is True:
            calibration = load_and_check_cal_info(calibration)
            gyro = calibration.calibrate_gyro(s.gyro.data)
            s.gyro.data = gyro
            s.gyro.is_calibrated = True
            s.gyro._unit = calibration.GYRO_UNIT
        return s

    def factory_calibrate_imu(self: T, inplace: bool = False) -> T:
        """Apply a calibration to the Acc and Gyro datastreams.

        The values used for that are taken from the datasheet of the sensor_type and are likely not to be accurate.
        For any tasks requiring precise sensor_type outputs, `calibrate_imu` should be used with measured calibration
        values.

        The final units of the output will be "g" for the Acc and "dps" (degrees per second) for the Gyro.

        Parameters
        ----------
        inplace :
            If True this methods modifies the current dataset object. If False, a copy of the dataset and all
            datastream objects is created
            Notes:
            This just combines `factory_calibrate_acc` and `factory_calibrate_gyro`.

        """
        s = self.factory_calibrate_acc(inplace=inplace)
        s = s.factory_calibrate_gyro(inplace=True)

        return s

    def factory_calibrate_gyro(self: T, inplace: bool = False) -> T:
        """Apply a factory calibration to the Gyro datastream.

        The values used for that are taken from the datasheet of the sensor_type and are likely not to be accurate.
        For any tasks requiring precise sensor_type outputs, `calibrate_gyro` should be used with measured calibration
        values.

        The final units of the output will be "dps" (degrees per second) for the Gyro.

        Parameters
        ----------
        inplace :
            If True this methods modifies the current dataset object. If False, a copy of the dataset and all
            datastream objects is created

        """
        s = inplace_or_copy(self, inplace)
        if self._check_calibration(s.gyro, "gyro") is True:
            s.gyro.data /= 2 ** 16 / self.info.gyro_range_dps / 2
            s.gyro.is_calibrated = True
        return s

    def factory_calibrate_acc(self: T, inplace: bool = False) -> T:
        """Apply a factory calibration to the Acc datastream.

        The values used for that are taken from the datasheet of the sensor_type and are likely not to be accurate.
        For any tasks requiring precise sensor_type outputs, `calibrate_acc` should be used with measured calibration
        values.

        The final units of the output will be "g" for the Acc.

        Parameters
        ----------
        inplace :
            If True this methods modifies the current dataset object. If False, a copy of the dataset and all
            datastream objects is created

        """
        s = inplace_or_copy(self, inplace)
        if self._check_calibration(s.acc, "acc") is True:
            s.acc.data /= 2 ** 16 / self.info.acc_range_g / 2
            s.acc.is_calibrated = True
        return s

    def factory_calibrate_baro(self: T, inplace: bool = False) -> T:
        """Apply a calibration to the Baro datastream.

        The values used for that are taken from the datasheet of the sensor_type and are likely not to be accurate.
        In general, if baro measurements are used to estimate elevation, the estimation should be calibrated relative to
        a reference altitude.

        The final units of the output will be "millibar" (equivalent to Hectopacal) for the Baro.

        Parameters
        ----------
        inplace :
            If True this methods modifies the current dataset object. If False, a copy of the dataset and all
            datastream objects is created

        """
        s = inplace_or_copy(self, inplace)
        if self._check_calibration(s.baro, "baro") is True:
            s.baro.data = (s.baro.data + 101325) / 100.0
            s.baro.is_calibrated = True
        return s

    def factory_calibrate_temperature(self: Type[T], inplace: bool = False):
        """Apply a factory calibration to the temperature datastream.

        The values used for that are taken from the datasheet of the sensor_type

        The final unit is Celsius.

        Parameters
        ----------
        inplace :
            If True this methods modifies the current dataset object. If False, a copy of the dataset and all
            datastream objects is created

        """
        s = inplace_or_copy(self, inplace)
        if self._check_calibration(s.temperature, "temperature") is True:
            s.temperature.data = s.temperature.data * (2 ** -9) + 23
            s.temperature.is_calibrated = True
        return s

    @staticmethod
    def _check_calibration(ds: Optional[Datastream], name: str):
        """Check if a specific datastream is already marked as calibrated, or if the datastream does not exist.

        In case the datastream is already calibrated a `RepeatedCalibrationError` is raised.
        In case the datastream does not exist, a warning is raised.

        Parameters
        ----------
        ds :
            datastream object or None
        name :
            name of the datastream object. Used to provide additional info in error messages.
        ds: Optional[Datastream] :

        """
        if ds is not None:
            if ds.is_calibrated is True:
                raise RepeatedCalibrationError(name)
            return True
        datastream_does_not_exist_warning(name, "calibration")
        return False

    def downsample(self: T, factor: int, inplace: bool = False) -> T:
        """Downsample all datastreams by a factor.

        This applies `scipy.signal.decimate` to all datastreams and the counter of the dataset.
        See :py:meth:`nilspodlib.datastream.Datastream.downsample` for details.

        .. warning::
            This will not modify any values in the header/info the dataset. I.e. the number of samples in the header/
            sync index values. Using methods that rely on these values might result in unexpected behaviour.
            For example `cut_to_syncregion` will not work correctly, if `cut`, `cut_counter_val`, or `downsample` was
            used before.

        Parameters
        ----------
        factor :
            Factor by which the dataset should be downsampled.
        inplace :
            If True this methods modifies the current dataset object. If False, a copy of the dataset and all
            datastream objects is created

        """
        s = inplace_or_copy(self, inplace)
        for key, val in s.datastreams:
            setattr(s, key, val.downsample(factor))
        s.counter = resample(s.counter, len(s.counter) // factor, axis=0)
        return s

    def cut(
        self: T,
        start: Optional[int] = None,
        stop: Optional[int] = None,
        step: Optional[int] = None,
        inplace: bool = False,
    ) -> T:
        """Cut all datastreams of the dataset.

        This is equivalent to applying the following slicing to all datastreams and the counter: array[start:stop:step]

        .. warning ::
            This will not modify any values in the header/info the dataset. I.e. the number of samples in the header/
            sync index values. Using methods that rely on these values might result in unexpected behaviour.
            For example `cut_to_syncregion` will not work correctly, if `cut` or `cut_counter_val` was used before.

        Parameters
        ----------
        start :
            Start index
        stop :
            Stop index
        step :
            Step size of the cut
        inplace :
            If True this methods modifies the current dataset object. If False, a copy of the dataset and all
            datastream objects is created

        """
        s = inplace_or_copy(self, inplace)

        for key, val in s.datastreams:
            setattr(s, key, val.cut(start, stop, step))
        sl = slice(start, stop, step)
        s.counter = s.counter[sl]
        return s

    def cut_counter_val(
        self: T,
        start: Optional[int] = None,
        stop: Optional[int] = None,
        step: Optional[int] = None,
        inplace: bool = False,
    ) -> T:
        """Cut the dataset based on values in the counter and not the index.

        Instead of just cutting the datastream based on its index, it is cut based on the counter value.
        This is equivalent to applying the following pandas style slicing to all datastreams and the counter:
        array.loc[start:stop:step]

        .. warning::
            This will not modify any values in the header/info the dataset. I.e. the number of samples in the header/
            sync index values. Using methods that rely on these values might result in unexpected behaviour.
            For example `cut_to_syncregion` will not work correctly, if `cut` or `cut_counter_val` was used before.

        Notes
        -----
        The method searches the respective index for the start and the stop value in the `counter` and calls `cut`
        with these values.
        The step size will be passed directly and not modified (i.e. the step size will not respect downsampling or
        similar operations done beforehand).

        Parameters
        ----------
        start :
            Start value in counter
        stop :
            Stop value in counter
        step :
            Step size of the cut
        inplace :
            If True this methods modifies the current dataset object. If False, a copy of the dataset and all
            datastream objects is created

        """
        if start:
            if start < self.counter[0]:
                raise ValueError("{} out of bounds for counter starting at {}".format(start, self.counter))
            start = np.searchsorted(self.counter, start)
        if stop:
            if stop > self.counter[-1]:
                raise ValueError("{} out of bounds for counter ending at {}".format(start, self.counter))
            stop = np.searchsorted(self.counter, stop)
        return self.cut(start, stop, step, inplace=inplace)

    def cut_to_syncregion(
        self: Type[T], start: bool = True, end: bool = False, warn_thres: Optional[int] = 30, inplace: bool = False
    ) -> T:
        """Cut the dataset to the region indicated by the first and last sync package received from master.

        This cuts the dataset to the values indicated by `info.sync_index_start` and `info.sync_index_stop`.
        In case the dataset was a sync-master (`info.sync_role = 'master'`) this will have no effect and the dataset
        will be returned unmodified.

        .. warning::
            This function should not be used after any other methods that can modify the counter (e.g. `cut` or
            `downsample`).

        .. warning::
            This will not modify any values in the header/info the dataset. I.e. the number of samples in the header/
            sync index values. Using methods that rely on these values might result in unexpected behaviour.

        Notes
        -----
        Usually to work with multiple syncronised datasets, a `SyncedSession` should be used instead of cutting
        the datasets manually. `SyncedSession.cut_to_syncregion` will cover multiple edge cases involving multiple
        datasets, which can not be handled by this method.


        Parameters
        ----------
        start :
            Whether the dataset should be cut at the `info.sync_index_start`.
            If this is False, a jump in the counter will remain.
            The only usecase for not cutting at the start is when the counters are already perfectly aligned.
        end :
            Whether the dataset should be cut at the `info.sync_index_stop`. Usually it can be assumed that the
            data will be synchronous for multiple seconds after the last sync package. Therefore, it might be
            acceptable to just ignore the last syncpackage and just cut the start of the dataset.
        warn_thres :
            Threshold in seconds from the end of a dataset. If the last syncpackage occurred more than
            warn_thres before the end of the dataset, a warning is emitted. Use warn_thres = None to silence.
            This is not relevant if the end of the dataset is cut (e.g. `end=True`)
        inplace :
            If True this methods modifies the current dataset object. If False, a copy of the dataset and all
            datastream objects is created

        Raises
        ------
        ValueError
            If the dataset does not have any sync infos
        ValueError
            If the dataset does not have any sync infos

        Warnings
        --------
        UserWarning
            If a syncpackage occurred far before the last sample in the dataset. See arg `warn_thres`.

        """
        if self.info.is_synchronised is False:
            raise ValueError("Only synchronised Datasets can be cut to the syncregion")
        if self.info.sync_role == "master":
            return inplace_or_copy(self, inplace)
        if warn_thres is not None and end is not True and self._check_sync_packages(warn_thres) is False:
            warnings.warn(
                "The last sync package occured more than {} s before the end of the measurement."
                "The last region of the data should not be trusted.".format(warn_thres),
                SynchronisationWarning,
            )
        end = self.info.sync_index_stop if end is True else None
        start = self.info.sync_index_start if start is True else None
        return self.cut(start, end, inplace=inplace)

    def data_as_df(
        self,
        datastreams: Optional[Sequence[str]] = None,
        index: Optional[str] = None,
        include_units: Optional[bool] = False,
    ) -> pd.DataFrame:
        """Export the datastreams of the dataset in a single pandas DataFrame.

        Parameters
        ----------
        datastreams :
            Optional list of datastream names, if only specific ones should be included. Datastreams that
            are not part of the current dataset will be silently ignored.
        index :
            Specify which index should be used for the dataset. The options are:
            "counter": For the actual counter
            "time": For the time in seconds since the first sample
            "utc": For the utc time stamp of each sample
            "utc_datetime": for a pandas DateTime index in UTC time
            None: For a simple index (0...N)
        include_units :
            If True the column names will have the unit of the datastream concatenated with an `_`
            Notes:
        include_units :
            If True the column names will have the unit of the datastream concatenated with an `_`
            Notes:
            This method calls the `data_as_df` methods of each Datastream object and then concats the results.
        include_units :
            If True the column names will have the unit of the datastream concatenated with an `_`

        Notes
        -----
        This method calls the `data_as_df` methods of each Datastream object and then concats the results.
        Therefore, it will use the column information of each datastream.

        Raises
        ------
        ValueError
            If any other than the allowed `index` values are used.

        """
        index_names = {None: "n_samples", "counter": "n_samples", "time": "t", "utc": "utc", "utc_datetime": "date"}
        if index and index not in index_names.keys():
            raise ValueError(
                "Supplied value for index ({}) is not allowed. Allowed values: {}".format(index, index_names.keys())
            )

        index_name = index_names[index]

        datastreams = datastreams or self.active_sensors
        dfs = [s.data_as_df(include_units=include_units) for k, s in self.datastreams if k in datastreams]

        df = pd.concat(dfs, axis=1)

        if index:
            if index != "counter":
                index += "_counter"
            index = getattr(self, index, None)
            df.index = index
        else:
            df = df.reset_index(drop=True)
        df.index.name = index_name
        return df

    def imu_data_as_df(self, index: Optional[str] = None, include_units: Optional[bool] = False) -> pd.DataFrame:
        """Export the acc and gyro datastreams of the dataset in a single pandas DataFrame.

        See Also
        --------
        nilspodlib.dataset.Dataset.data_as_df

        Parameters
        ----------
        index :
            Specify which index should be used for the dataset. The options are:
            "counter": For the actual counter
            "time": For the time in seconds since the first sample
            "utc": For the utc time stamp of each sample
            "utc_datetime": for a pandas DateTime index in UTC time
            None: For a simple index (0...N)
        include_units :
            If True the column names will have the unit of the datastream concatenated with an `_`
            Notes:
        include_units :
            If True the column names will have the unit of the datastream concatenated with an `_`
            Notes:
            This method calls the `data_as_df` methods of each Datastream object and then concats the results.
        include_units :
            If True the column names will have the unit of the datastream concatenated with an `_`

        Notes
        -----
        This method calls the `data_as_df` methods of each Datastream object and then concats the results.
        Therefore, it will use the column information of each datastream.


        Raises
        ------
        ValueError
            If any other than the allowed `index` values are used.

        """
        return self.data_as_df(datastreams=["acc", "gyro"], index=index, include_units=include_units)

    def find_calibrations(
        self,
        folder: Optional[path_t] = None,
        recursive: bool = True,
        filter_cal_type: Optional[str] = None,
        ignore_file_not_found: Optional[bool] = False,
    ) -> List[Path]:
        """Find all calibration infos that belong to a given sensor_type.

        As this only checks the filenames, this might return a false positive depending on your folder structure and
        naming.

        Parameters
        ----------
        folder :
            Basepath of the folder to search. If None, tries to find a default calibration
        recursive :
            If the folder should be searched recursive or not.
        filter_cal_type :
            Whether only files obtain with a certain calibration type should be found.
            This will look for the `CalType` inside the json file and hence cause performance problems.
            If None, all found files will be returned.
            For possible values, see the `imucal` library.
        ignore_file_not_found :
            If True this function will not raise an error, but rather return an empty list, if no
            calibration files were found for the specific sensor_type.

        See Also
        --------
        nilspodlib.calibration_utils.find_calibrations_for_sensor

        """
        # TODO: Test
        return find_calibrations_for_sensor(
            sensor_id=self.info.sensor_id,
            folder=folder,
            recursive=recursive,
            filter_cal_type=filter_cal_type,
            ignore_file_not_found=ignore_file_not_found,
        )

    def find_closest_calibration(
        self,
        folder: Optional[path_t] = None,
        recursive: bool = True,
        filter_cal_type: Optional[str] = None,
        before_after: Optional[str] = None,
        warn_thres: datetime.timedelta = datetime.timedelta(days=30),  # noqa E252
        ignore_file_not_found: Optional[bool] = False,
    ) -> Path:
        """Find the closest calibration info to the start of the measurement.

        As this only checks the filenames, this might return a false positive depending on your folder structure and
        naming.

        Parameters
        ----------
        folder :
            Basepath of the folder to search. If None, tries to find a default calibration
        recursive :
            If the folder should be searched recursive or not.
        filter_cal_type :
            Whether only files obtain with a certain calibration type should be found.
            This will look for the `CalType` inside the json file and hence cause performance problems.
            If None, all found files will be returned.
            For possible values, see the `imucal` library.
        before_after :
            Can either be 'before' or 'after', if the search should be limited to calibrations that were
            either before or after the specified date.
        warn_thres :
            If the distance to the closest calibration is larger than this threshold, a warning is emitted
        ignore_file_not_found :
            If True this function will not raise an error, but rather return `None`, if no
            calibration files were found for the specific sensor_type.

        See Also
        --------
        nilspodlib.calibration_utils.find_calibrations_for_sensor
        nilspodlib.calibration_utils.find_closest_calibration_to_date

        """
        # TODO: Test
        return find_closest_calibration_to_date(
            sensor_id=self.info.sensor_id,
            cal_time=self.info.utc_datetime_start,
            folder=folder,
            recursive=recursive,
            filter_cal_type=filter_cal_type,
            before_after=before_after,
            warn_thres=warn_thres,
            ignore_file_not_found=ignore_file_not_found,
        )

    def _check_sync_packages(self, threshold_s: int = 30, where="end") -> bool:
        """Check if the last sync package occurred far from the actual end of the recording.

        This can be the case, if the master stopped sending packages, or if the sensor could not receive any new sync
        info for a prelonged period of time.
        In particular in the latter case, careful review of the data is advised.
        """
        if self.info.sync_role == "slave":
            if where == "end":
                tmp = len(self.counter) - self.info.sync_index_stop
            elif where == "start":
                tmp = self.info.sync_index_start
            else:
                raise ValueError('Invalid value for "where" encountered.')
            if tmp > threshold_s * self.info.sampling_rate_hz:
                return False
        return True


def parse_binary(path: path_t, legacy_support: str = "error") -> Tuple[Dict[str, np.ndarray], np.ndarray, Header]:
    """Parse a binary NilsPod session file and read the header and the data.

    Parameters
    ----------
    path :
        Path to the file
    legacy_support :
        This indicates how to deal with old firmware versions.
        If `error`, An error is raised, if an unsupported version is detected.
        If `warn`, A warning is raised, but the file is parsed without modification
        If `resolve`, A legacy conversion is performed to load old files. If no suitable conversion is found,
        an error is raised. See the `legacy` package and the README to learn more about available conversions.

    Returns
    -------
    sensor_data :
        The sensor data as dictionary
    counter :
        The counter values
    session_header :
        The session header

    Raises
    ------
    VersionError
        If unsupported FirmwareVersion is detected and `legacy_error` is `error`
    VersionError
        If `legacy_error` is `resolve`, but no suitable conversion is found.

    """
    header_bytes, data_bytes = get_header_and_data_bytes(path)

    version = get_strict_version_from_header_bytes(header_bytes)

    if legacy_support == "resolve":
        header_bytes, data_bytes = find_conversion_function(version, in_memory=True)(header_bytes, data_bytes)
    elif legacy_support in ["error", "warn"]:
        legacy_support_check(version, as_warning=(legacy_support == "warn"))
    else:
        raise ValueError("legacy_support must be one of 'resolve',  'error' or 'warn'")

    session_header = Header.from_bin_array(header_bytes[1:])

    sample_size = session_header.sample_size
    n_samples = session_header.n_samples

    data = read_binary_uint8(data_bytes, sample_size, n_samples)

    counter, sensor_data = split_into_sensor_data(data, session_header)

    if session_header.strict_version_firmware >= StrictVersion("0.13.0") and len(counter) != session_header.n_samples:
        warnings.warn(
            "The number of samples in the dataset does not match the number indicated by the header."
            "This might indicate a corrupted file",
            LegacyWarning,
        )

    return sensor_data, counter, session_header


def split_into_sensor_data(data: np.ndarray, session_header: Header) -> Tuple[np.ndarray, Dict[str, np.ndarray]]:
    """Split/Parse the binary data into the different sensors and the counter."""
    sensor_data = dict()

    idx = 0
    for sensor in session_header.enabled_sensors:
        bits, channel, dtype = SENSOR_SAMPLE_LENGTH[sensor]
        bits_per_channel = bits // channel
        tmp = np.full((len(data), channel), np.nan)
        for i in range(channel):
            tmp[:, i] = convert_little_endian(
                np.atleast_2d(data[:, idx : idx + bits_per_channel]).T, dtype=dtype
            ).astype(float)
            idx += bits_per_channel
        sensor_data[sensor] = tmp

    len_counter, _, counter_dtype = SENSOR_SAMPLE_LENGTH["counter"]

    # Sanity Check:
    if not idx + len_counter == data.shape[-1]:
        expected_cols = idx
        all_cols = data.shape[-1] - len_counter
        raise InvalidInputFileError(
            "The input file has an invalid format. {} data columns expected based on the header, but {} exist.".format(
                expected_cols, all_cols
            )
        )

    counter = convert_little_endian(np.atleast_2d(data[:, -len_counter:]).T, dtype=counter_dtype)

    return counter, sensor_data



"""Internal bases for sessions to make it easier to call dataset methods on the session object."""

from functools import wraps
from typing import Optional, Iterable, Tuple, TypeVar, TYPE_CHECKING, Type, Sequence

import numpy as np

from nilspodlib.dataset import Dataset
from nilspodlib.utils import path_t, inplace_or_copy, remove_docstring_indent

T = TypeVar("T")

if TYPE_CHECKING:
    from nilspodlib.datastream import Datastream  # noqa: F401
    import pandas as pd  # noqa: F401


class CascadingDatasetField:
    """A simple descriptor object to forward attribute access to all datasets of a session."""

    name: str
    __doc__: str

    def __set_name__(self, owner, name):
        """Set the name of the field and update the docstring, by pulling from the Dataset class."""
        self.name = name
        self.__doc__ = getattr(Dataset, self.name, None).__doc__

    def __get__(self, instance, owner):
        """Get the attribute from all nested objects."""
        return tuple(getattr(d, self.name) for d in instance.datasets)


def call_dataset(autogen_doc=True):  # noqa: D202
    """Forward all method calls to all datasets of a session.

    This function respects the inplace feature and will create a copy of the session object if required.

    Parameters
    ----------
    autogen_doc :
        If True, the docstring of the respective dataset method is copied to the method with short pretext.
        If a docstring already exists, the dataset docstring will be appended WITHOUT pretext. (Default value = True)

    """

    def _wrapped(method):
        @wraps(method)
        def _cascading_access(*args, **kwargs):
            session = args[0]
            return_vals = tuple(getattr(d, method.__name__)(*args[1:], **kwargs) for d in session.datasets)

            if all(isinstance(d, Dataset) for d in return_vals):
                inplace = kwargs.get("inplace", False)
                s = inplace_or_copy(session, inplace)
                s.datasets = return_vals
                return s
            return return_vals

        if autogen_doc:
            if _cascading_access.__doc__:
                _cascading_access.__doc__ += "\n\n"
            else:
                _cascading_access.__doc__ = (
                    "Apply `Dataset.{0}` to all datasets of the session.\n\n"
                    "See :py:meth:`nilspodlib.dataset.Dataset.{0}` for more details. "
                    "The docstring of this method is included below:\n\n".format(method.__name__)
                )
            _cascading_access.__doc__ += remove_docstring_indent(getattr(Dataset, method.__name__).__doc__)
        return _cascading_access

    return _wrapped


class _MultiDataset:
    """Wrapper that holds all attributes and methods that can be simply called on multiple datasets.

    Notes
    -----
    This class should not be used as public interface and is only relevant as base for the session class

    This class uses a decorator for methods and a descriptor for attributes to automatically forward all calls to
    multiple datasets.
    See the implementation of `CascadingDatasetField` and `call_dataset` for details.

    """

    path: path_t = CascadingDatasetField()
    acc: Tuple[Optional["Datastream"]] = CascadingDatasetField()
    gyro: Tuple[Optional["Datastream"]] = CascadingDatasetField()
    mag: Tuple[Optional["Datastream"]] = CascadingDatasetField()
    baro: Tuple[Optional["Datastream"]] = CascadingDatasetField()
    analog: Tuple[Optional["Datastream"]] = CascadingDatasetField()
    ecg: Tuple[Optional["Datastream"]] = CascadingDatasetField()
    ppg: Tuple[Optional["Datastream"]] = CascadingDatasetField()
    temperature: Tuple[Optional["Datastream"]] = CascadingDatasetField()
    counter: Tuple[np.ndarray] = CascadingDatasetField()

    size: Tuple[int] = CascadingDatasetField()
    datastreams: Tuple[Iterable["Datastream"]] = CascadingDatasetField()

    ACTIVE_SENSORS: Tuple[Tuple[str]] = CascadingDatasetField()

    # This needs to be implemented by the session
    datasets: Tuple[Dataset]

    @call_dataset()
    def factory_calibrate_imu(self: Type[T], inplace: bool = False) -> T:  # noqa: D105
        pass

    @call_dataset()
    def factory_calibrate_gyro(self: Type[T], inplace: bool = False) -> T:  # noqa: D105
        pass

    @call_dataset()
    def factory_calibrate_baro(self: Type[T], inplace: bool = False) -> T:  # noqa: D105
        pass

    @call_dataset()
    def factory_calibrate_temperature(self: Type[T], inplace: bool = False) -> T:  # noqa: D105
        pass

    @call_dataset()
    def cut_to_syncregion(  # noqa: D105
        self: Type[T], start: bool = True, end: bool = False, warn_thres: Optional[int] = 30, inplace: bool = False
    ) -> T:
        pass

    @call_dataset()
    def cut(  # noqa: D105
        self: Type[T],
        start: Optional[int] = None,
        stop: Optional[int] = None,
        step: Optional[int] = None,
        inplace: bool = False,
    ) -> T:
        pass

    @call_dataset()
    def cut_counter_val(  # noqa: D105
        self: Type[T],
        start: Optional[int] = None,
        stop: Optional[int] = None,
        step: Optional[int] = None,
        inplace: bool = False,
    ) -> T:
        pass

    @call_dataset()
    def downsample(self: Type[T], factor: int, inplace: bool = False) -> T:  # noqa: D105
        pass

    @call_dataset()
    def data_as_df(  # noqa: D105
        self,
        datastreams: Optional[Sequence[str]] = None,
        index: Optional[str] = None,
        include_units: Optional[bool] = True,
    ) -> Tuple["pd.DataFrame"]:
        pass

    @call_dataset()
    def imu_data_as_df(self, index: Optional[str] = None) -> Tuple["pd.DataFrame"]:  # noqa: D105
        pass

    @call_dataset()
    def find_closest_calibration(  # noqa: D105
        self,
        folder: Optional[path_t] = None,
        recursive: bool = False,
        filter_cal_type: Optional[str] = None,
        before_after: Optional[str] = None,
        ignore_file_not_found: Optional[bool] = False,
    ):
        pass

    @call_dataset()
    def find_calibrations(  # noqa: D105
        self,
        folder: Optional[path_t] = None,
        recursive: bool = True,
        filter_cal_type: Optional[str] = None,
        ignore_file_not_found: Optional[bool] = False,
    ):
        pass



"""Legacy support helper to convert older NilsPod files into new versions."""
import warnings
from distutils.version import StrictVersion
from typing import Tuple, Callable, Optional, Union

import numpy as np

from nilspodlib.exceptions import LegacyWarning, VersionError
from nilspodlib.utils import (
    path_t,
    get_header_and_data_bytes,
    get_strict_version_from_header_bytes,
    get_sample_size_from_header_bytes,
)

CONVERSION_DICT = {
    "12_0": {"min": StrictVersion("0.11.255"), "max": StrictVersion("0.13.255")},
    "11_2": {"min": StrictVersion("0.11.2"), "max": StrictVersion("0.11.255")},
}

MIN_NON_LEGACY_VERSION = StrictVersion("0.14.0")


def find_conversion_function(
    version: StrictVersion, in_memory: Optional[bool] = True, return_name: Optional[bool] = False
) -> Union[Callable, str]:
    """Find a method that is able to convert a recording from one version to the other.

    This will either return one of the `load_{}` functions, if `in_memory` is True or the `convert_{}` variant if False.
    """
    if version >= MIN_NON_LEGACY_VERSION:
        return lambda x, y: (x, y)

    for k, v in CONVERSION_DICT.items():
        if v["min"] <= version < v["max"]:
            n = "load_" if in_memory else "convert_"
            if return_name:
                return n + k
            return globals()[n + k]
    raise VersionError("No suitable conversion function found for {}".format(version))


def convert_12_0(in_path: path_t, out_path: path_t) -> None:
    """Convert a session recorded with a firmware version >0.11.255 and <0.13.255 to the most up-to-date format.

    This will update the firmware version to 0.13.255 to identify converted sessions.
    Potential version checks in the library will use <0.13.255 to check for compatibility.

    .. warning:: After the update the following features will not work:
                    - The sync group was removed and hence can not be read anymore

    Parameters
    ----------
    in_path :
        path to 0.12.x / 0.13.x file
    out_path :
        path to converted 0.13.255 file

    """
    header, data_bytes = get_header_and_data_bytes(in_path)
    header, data_bytes = load_12_0(header, data_bytes)

    with open(out_path, "wb+") as f:
        f.write(bytearray(header))
        f.write(bytearray(data_bytes))


def load_12_0(header: np.ndarray, data_bytes: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Convert a session recorded with a firmware version >0.11.255 and <0.13.255 to the most up-to-date format.

    This will update the firmware version to 0.13.255 to identify converted sessions.
    Potential version checks in the library will use <0.13.255 to check for compatibility.

    .. warning:: After the update the following features will not work:
                    - The sync group was removed and hence can not be read anymore

    Parameters
    ----------
    header :
        bytes containing all the legacy header information
    data_bytes :
        raw bytes representing the data

    """
    min_v = CONVERSION_DICT["12_0"]["min"]
    max_v = CONVERSION_DICT["12_0"]["max"]
    version = get_strict_version_from_header_bytes(header)

    if not min_v <= version < max_v:
        raise VersionError(
            "This converter is meant for files recorded with Firmware version after {} and before {}"
            " not {}".format(min_v, max_v, version)
        )

    header = _shift_bytes_12_0(header)

    # Update firmware version
    header[-2] = 13
    header[-1] = 255

    return header, data_bytes


def convert_11_2(in_path: path_t, out_path: path_t) -> None:
    """Convert a session recorded with a 0.11.<2 firmware to the most up-to-date format.

    This will update the firmware version to 0.13.255 to identify converted sessions.
    Potential version checks in the library will use <0.13.255 to check for compatibility.

    .. warning:: After the update the following features will not work:
                    - The battery sensor_type does not exist anymore and hence, is not supported in the converted files
                    - The sync group was removed and hence can not be read anymore

    Parameters
    ----------
    in_path :
        path to 0.11.2 file
    out_path :
        path to converted 0.13.255 file

    """
    header, data_bytes = get_header_and_data_bytes(in_path)
    header, data_bytes = load_11_2(header, data_bytes)

    with open(out_path, "wb+") as f:
        f.write(bytearray(header))
        f.write(bytearray(data_bytes))


def load_11_2(header: np.ndarray, data_bytes: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Convert a session recorded with a 0.11.<2 firmware to the most up-to-date format.

    This will update the firmware version to 0.13.255 to identify converted sessions.
    Potential version checks in the library will use <0.13.255 to check for compatibility.

    .. warning:: After the update the following features will not work:
                    - The battery sensor_type does not exist anymore and hence, is not supported in the converted files
                    - The sync group was removed and hence can not be read anymore

    Parameters
    ----------
    header :
        Bytes containing all the legacy header information
    data_bytes :
        Raw bytes representing the data

    """
    min_v = CONVERSION_DICT["11_2"]["min"]
    max_v = CONVERSION_DICT["11_2"]["max"]
    version = get_strict_version_from_header_bytes(header)

    if not min_v <= version < max_v:
        raise VersionError(
            "This converter is meant for files recorded with Firmware version after {} and before {}"
            " not {}".format(min_v, max_v, version)
        )

    packet_size = get_sample_size_from_header_bytes(header)

    data_bytes = _fix_little_endian_counter(data_bytes, packet_size).flatten()

    header = _insert_missing_bytes_11_2(header)
    header[3:5] = _split_sampling_rate_byte_11_2(header[3])
    header[2] = _convert_sensor_enabled_flag_11_2(header[2])

    # adapt to new header size:
    header[0] = len(header)

    # Update firmware version
    header[-1] = 255

    header, data_bytes = load_12_0(header, data_bytes)
    return header, data_bytes


def _fix_little_endian_counter(data_bytes, packet_size):
    """Convert the big endian counter of older firmware versions to a small endian counter."""
    expected_samples = len(data_bytes) // packet_size
    data_bytes = data_bytes[: expected_samples * packet_size]
    data = np.reshape(data_bytes, (expected_samples, int(packet_size)))
    data[:, -4:] = np.flip(data[:, -4:], axis=-1)
    return data


def _convert_sensor_enabled_flag_11_2(byte):
    """Convert the old sensor enabled flags to the new one."""
    conversion_map = {0x01: 0x02, 0x02: 0x10, 0x04: 0x08, 0x08: 0x80}  # gyro  # analog  # baro  # temperature

    # always enable acc for old sessions:
    out_byte = 0x01

    # convert other sensors if enabled
    for old, new in conversion_map.items():
        if bool(byte & old) is True:
            out_byte |= new
    return out_byte


def _insert_missing_bytes_11_2(header_bytes):
    """Insert header bytes that were added after 11.2."""
    header_bytes = np.insert(header_bytes, 4, 0x00)

    header_bytes = np.insert(header_bytes, 47, [0x00] * 2)

    return header_bytes


def _shift_bytes_12_0(header_bytes):
    """Move old bytestructure, as it was changed for versions after 12.0."""
    # remove old sync_group byte:
    header_bytes = np.delete(header_bytes, 7)

    # Add new empty byte after enabled sensors
    header_bytes = np.insert(header_bytes, 3, 0x00)

    return header_bytes


def _split_sampling_rate_byte_11_2(sampling_rate_byte: int) -> Tuple[int, int]:
    """Separate sampling rate into its own byte."""
    return sampling_rate_byte & 0x0F, sampling_rate_byte & 0xF0


def legacy_support_check(version: StrictVersion, as_warning: bool = False):
    """Check if a file recorded with a specific fileformat version can be converted using legacy support.

    Parameters
    ----------
    version :
        The version to check for.
    as_warning :
        If True only a Warning instead of an error is raised, if legacy support is required for the dataset.

    """
    if version < StrictVersion("0.11.2"):
        msg = "You are using a version ({}) previous to 0.11.2. This version is not supported!".format(version)
    elif version >= StrictVersion("0.13.255"):
        return
    else:
        try:
            converter = find_conversion_function(version, in_memory=False, return_name=True)
            msg = (
                "You are using a version ({}) which is only supported by legacy support."
                " Use `{}` to update the binary format to a newer version"
                ' or use `legacy_support="resolve"` when loading the file'.format(version, converter)
            )
        except VersionError:
            msg = "You are using a version completely unknown version: {}".format(version)

    if as_warning is True:
        warnings.warn(msg, LegacyWarning)
    else:
        raise VersionError(msg)



# -*- coding: utf-8 -*-
"""Session groups multiple Datasets from sensors recorded at the same time."""
import datetime
import warnings
from pathlib import Path
from typing import Iterable, Tuple, TypeVar, Type, Optional, Union, TYPE_CHECKING, Sequence

import numpy as np

from nilspodlib._session_base import _MultiDataset
from nilspodlib.dataset import Dataset
from nilspodlib.exceptions import SynchronisationError, SynchronisationWarning
from nilspodlib.header import _ProxyHeader
from nilspodlib.utils import validate_existing_overlap, inplace_or_copy, path_t

if TYPE_CHECKING:
    from imucal import CalibrationInfo  # noqa: F401
    import pandas as pd  # noqa: F401

T = TypeVar("T", bound="Session")


class Session(_MultiDataset):
    """Object representing a collection of Datasets.

    .. note:: By default a session makes no assumption about when and how datasets are recorded.
              It just provides an interface to manipulate multiple datasets at once.
              If you have datasets that were recorded simultaneously with active sensor_type synchronisation,
              you should use a `SyncedSession` instead of a `Session` to take full advantage of this.

    A session can access all the same attributes and most of the methods provided by a dataset.
    However, instead of returning a single value/acting only on a single dataset, it will return a tuple of values (one
    for each dataset) or modify all datasets of a session.
    You can also use the `self.info` object to access header information of all datasets at the same time.
    All return values will be in the same order as `self.datasets`.

    Attributes
    ----------
    datasets
        A tuple of the datasets belonging to the session
    info
        Get the metadata (Header) information for all datasets

    See Also
    --------
    nilspodlib.session.SyncedSession

    """

    datasets: Tuple[Dataset]

    @property
    def info(self):
        """Get metainfo for all datasets.

        All attributes of :py:class:`nilspodlib.header.HeaderInfo` are supported in read-only mode.
        """
        return _ProxyHeader(tuple(d.info for d in self.datasets))

    def __init__(self, datasets: Iterable[Dataset]):
        """Create new session.

        Instead of this init you can also use the factory methods :py:meth:`~nilspodlib.session.Session.from_file_paths`
        and :py:meth:`~nilspodlib.session.Session.from_folder_path`.

        Parameters
        ----------
        datasets :
            List of :py:class:`nilspodlib.dataset.Dataset` instances, which should be grouped into a session.

        """
        self.datasets = tuple(datasets)

    @classmethod
    def from_file_paths(cls: Type[T], paths: Iterable[path_t], legacy_support: str = "error") -> T:
        """Create a new session from a list of files pointing to valid .bin files.

        Parameters
        ----------
        paths :
            List of paths pointing to files to be included
        legacy_support :
            This indicates how to deal with old firmware versions.
            If `error`: An error is raised, if an unsupported version is detected.
            If `warn`: A warning is raised, but the file is parsed without modification
            If `resolve`: A legacy conversion is performed to load old files. If no suitable conversion is found,
            an error is raised. See the `legacy` package and the README to learn more about available
            conversions.

        """
        ds = (Dataset.from_bin_file(p, legacy_support=legacy_support) for p in paths)
        return cls(ds)

    @classmethod
    def from_folder_path(
        cls: Type[T], base_path: path_t, filter_pattern: str = "*.bin", legacy_support: str = "error"
    ) -> T:
        """Create a new session from a folder path containing valid .bin files.

        Parameters
        ----------
        base_path :
            Path to the folder
        filter_pattern :
            regex that can be used to filter the files in the folder. This is passed to Pathlib.glob()
        legacy_support :
            This indicates how to deal with old firmware versions.
            If `error`: An error is raised, if an unsupported version is detected.
            If `warn`: A warning is raised, but the file is parsed without modification
            If `resolve`: A legacy conversion is performed to load old files. If no suitable conversion is found,
            an error is raised. See the `legacy` package and the README to learn more about available
            conversions.

        """
        ds = list(Path(base_path).glob(filter_pattern))
        if not ds:
            raise ValueError('No files matching "{}" where found in {}'.format(filter_pattern, base_path))
        return cls.from_file_paths(ds, legacy_support=legacy_support)

    def get_dataset_by_id(self, sensor_id: str) -> Dataset:
        """Get a specific dataset by its sensor_type id.

        Parameters
        ----------
        sensor_id :
            Four letter/digit unique id of the sensor

        """
        return self.datasets[self.info.sensor_id.index(sensor_id)]

    def calibrate_imu(self: T, calibrations: Iterable[Union["CalibrationInfo", path_t]], inplace: bool = False) -> T:
        """Calibrate the imus of all datasets by providing a list of calibration infos.

        If you do not want to calibrate a specific IMU, you can pass `None` for its position.

        Parameters
        ----------
        calibrations :
            List of calibration infos in the same order than `self.datasets`
        inplace :
            If True this methods modifies the current session object. If False, a copy of the sesion and all
            dataset objects is created

        See Also
        --------
        nilspodlib.dataset.Dataset.calibrate_imu

        """
        s = inplace_or_copy(self, inplace)
        s.datasets = [d.calibrate_imu(c, inplace=True) if c else d for d, c in zip(s.datasets, calibrations)]
        return s

    def calibrate_acc(self: T, calibrations: Iterable[Union["CalibrationInfo", path_t]], inplace: bool = False) -> T:
        """Calibrate the accs of all datasets by providing a list of calibration infos.

        If you do not want to calibrate a specific IMU, you can pass `None` for its position.

        Parameters
        ----------
        calibrations :
            List of calibration infos in the same order than `self.datasets`
        inplace :
            If True this methods modifies the current session object. If False, a copy of the sesion and all
            dataset objects is created

        See Also
        --------
        nilspodlib.dataset.Dataset.calibrate_acc

        """
        s = inplace_or_copy(self, inplace)
        s.datasets = [d.calibrate_acc(c, inplace=True) if c else d for d, c in zip(s.datasets, calibrations)]
        return s

    def calibrate_gyro(self: T, calibrations: Iterable[Union["CalibrationInfo", path_t]], inplace: bool = False) -> T:
        """Calibrate the gyros of all datasets by providing a list of calibration infos.

        If you do not want to calibrate a specific IMU, you can pass `None` for its position.

        Parameters
        ----------
        calibrations :
            List of calibration infos in the same order than `self.datasets`
        inplace :
            If True this methods modifies the current session object. If False, a copy of the session and all
            dataset objects is created

        See Also
        --------
        nilspodlib.dataset.Dataset.calibrate_gyro

        """
        s = inplace_or_copy(self, inplace)
        s.datasets = [d.calibrate_gyro(c, inplace=True) if c else d for d, c in zip(s.datasets, calibrations)]
        return s


class SyncedSession(Session):
    """Object representing a collection of Datasets recorded with active synchronisation.

    A session can access all the same attributes and most of the methods provided by a dataset.
    However, instead of returning a single value/acting only on a single dataset, it will return a tuple of values (one
    for each dataset) or modify all datasets of a session.
    You can also use the `self.info` object to access header information of all datasets at the same time.
    All return values will be in the same order as `self.datasets`.

    To synchronise a dataset, you usually want to call :py:meth:`~nilspodlib.session.SyncedSession.cut_to_syncregion`
    on the session.
    The resulting session is considered fully synchronised (depending on the parameters chosen).
    This means that all datasets have the same length and the exact same counter.
    However, note, that the header information of the individual datasets will not be updated to reflect the sync.
    This means that header values like `number_of_samples` or the start and stop times will not match the data anymore.
    As a substitute you can use a set of direct attributes on the session (e.g. `session_utc_start`, `session_duration`,
    etc.)

    Attributes
    ----------
    datasets
        A tuple of the datasets belonging to the session
    info
        Get the metadata (Header) information for all datasets
    master
        Get the dataset belonging to the sync-server sensor.
    slaves
        Get the datasets belonging to all the sync-clients sensors.
    session_utc_start
       Start time of the session as utc timestamp.
       You need to `cut_to_sync_region` before you can obtain this value.
    session_utc_stop
       Stop time of the session as utc timestamp.
       You need to `cut_to_sync_region` before you can obtain this value.
    session_duration
        Duration of the session in seconds.
        You need to `cut_to_sync_region` before you can obtain this value.
    session_utc_datetime_start
        Start time of the session as utc datetime.
        You need to `cut_to_sync_region` before you can obtain this value.
    session_utc_datetime_stop
        Stop time of the session as utc datetime.
        You need to `cut_to_sync_region` before you can obtain this value.
    VALIDATE_ON_INIT
        If True all synced sessions will be checked on init.
        These checks include testing, if all datasets are really part of a single measurement.
        In rare cases, it might be useful to deactivate these checks and force the creation of a synced session.
        In this case you need to set this **class** attribute to false before loading the session:

        >>> SyncedSession.VALIDATE_ON_INIT = False
        >>> SyncedSession.from_folder_path('./my/path') # No validation will be performed

    See Also
    --------
    nilspodlib.session.Session

    """

    VALIDATE_ON_INIT: bool = True
    _fully_synced = False

    @property
    def master(self) -> Dataset:
        """Get the dataset belonging to the sync-server sensor."""
        return next(d for d in self.datasets if d.info.sync_role == "master")

    @property
    def slaves(self) -> Tuple[Dataset]:
        """Get the datasets belonging to all the sync-clients sensors."""
        return tuple(d for d in self.datasets if d.info.sync_role == "slave")

    @property
    def session_utc_start(self) -> float:
        """Start time of the session as utc timestamp."""
        return self.session_utc_datetime_start.timestamp()

    @property
    def session_utc_stop(self) -> float:
        """Stop time of the session as utc timestamp."""
        return self.session_utc_datetime_stop.timestamp()

    @property
    def session_duration(self) -> float:
        """Duration of the session in seconds."""
        return self.session_utc_stop - self.session_utc_start

    @property
    def session_utc_datetime_start(self) -> datetime.datetime:
        """Start time of the session as utc datetime."""
        if not self._fully_synced:
            raise SynchronisationError("Only fully synced datasets, have valid start and stop times.")
        return self.master.info.utc_datetime_start_day_midnight + datetime.timedelta(
            seconds=self.master.counter[0] / self.master.info.sampling_rate_hz
        )

    @property
    def session_utc_datetime_stop(self) -> datetime.datetime:
        """Stop time of the session as utc datetime."""
        if not self._fully_synced:
            raise SynchronisationError("Only fully synced datasets, have valid start and stop times.")
        return self.master.info.utc_datetime_start_day_midnight + datetime.timedelta(
            seconds=self.master.counter[-1] / self.master.info.sampling_rate_hz
        )

    def __init__(self, datasets: Iterable[Dataset]):
        """Create new synced session.

        Instead of this init you can also use the factory methods
        :py:meth:`~nilspodlib.session.SyncedSession.from_file_paths` and
        :py:meth:`~nilspodlib.session.SyncedSession.from_folder_path`.

        This init performs basic validation on the datasets.
        See :py:meth:`~nilspodlib.session.SyncedSession.validate` for details.

        Parameters
        ----------
        datasets :
            List of :py:class:`nilspodlib.dataset.Dataset` instances, which should be grouped into a session.

        """
        super().__init__(datasets)
        if self.VALIDATE_ON_INIT:
            self.validate()

    def validate(self) -> None:
        """Check if basic properties of a synced session are fulfilled.

        Raises
        ------
        ValueError
            This raises a ValueError in the following cases:
            - One or more of the datasets are not part of the same syncgroup/same sync channel
            - Multiple datasets are marked as "master"
            - One or more datasets indicate that they are not synchronised
            - One or more dataset has a different sampling rate than the others
            - If the recording times of provided datasets do not have any overlap

        """
        if not self._validate_sync_groups():
            raise ValueError("The provided datasets are not part of the same sync_group")
        master_valid, slaves_valid = self._validate_sync_role()
        if not master_valid:
            raise ValueError("SyncedSessions require exactly 1 master.")
        if not slaves_valid:
            raise ValueError("One of the provided sessions is not correctly set to either slave or master")
        if not self._validate_sampling_rate():
            raise ValueError("All provided sessions need to have the same sampling rate")
        if not self._validate_overlapping_record_time():
            raise ValueError("The provided datasets do not have any overlapping time period.")

    def _validate_sync_groups(self) -> bool:
        """Check that all _headers belong to the same sync group."""
        sync_channel = set(self.info.sync_channel)
        sync_address = set(self.info.sync_address)
        return all((len(i) == 1 for i in [sync_channel, sync_address]))

    def _validate_sync_role(self) -> Tuple[bool, bool]:
        """Check that there is only 1 master and all other sensors were configured as slaves."""
        roles = self.info.sync_role
        master_valid = len([i for i in roles if i == "master"]) == 1
        slaves_valid = len([i for i in roles if i == "slave"]) == len(roles) - 1
        return master_valid, slaves_valid

    def _validate_sampling_rate(self) -> bool:
        """Check that all sensors had the same sampling rate."""
        sr = set(self.info.sampling_rate_hz)
        return len(sr) == 1

    def _validate_overlapping_record_time(self) -> bool:
        """Check if all provided sessions have overlapping recording times."""
        start_times = np.array(self.info.utc_start)
        stop_times = np.array(self.info.utc_stop)
        return validate_existing_overlap(start_times, stop_times)

    def align_to_syncregion(
        self: Type[T],
        cut_start: bool = False,
        cut_end: bool = False,
        inplace: bool = False,
        warn_thres: Optional[int] = 30,
    ) -> T:
        """Align all datasets based on regions where they were synchronised to the master.

        At the end all datasets are cut to the same length, so that the maximum overlap between all datasets is
        preserved.

        Parameters
        ----------
        cut_start :
            Whether the dataset should be cut at `info.sync_index_start`.
            If `False` a new corrected counter value will be calculated for all packages before the first
            syncpackage.
            Usually it can be assumed that this extrapolation is valid for multiple seconds before the first
            package.
        cut_end :
            Whether the dataset should be cut at the `info.sync_index_stop`. Usually it can be assumed that the
            data will be synchronous for multiple seconds after the last sync package. Therefore, it might be
            acceptable to just ignore the last syncpackage and just cut the start of the dataset.
        warn_thres :
            Threshold in seconds from the end of a dataset. If the last syncpackage occurred more than
            warn_thres before the end of the dataset, a warning is emitted. Use warn_thres = None to silence.
            This is not relevant if the end of the dataset is cut (e.g. `end=True`)
        inplace :
            If operation should be performed on the current Session object, or on a copy
            Warns:
        inplace :
            If operation should be performed on the current Session object, or on a copy
            Warns:
            If a syncpackage occurred far before the last sample in any of the dataset. See arg `warn_thres`.

        """
        # Correct counter before first sync package
        s = inplace_or_copy(self, inplace)

        if s._fully_synced is True:
            raise SynchronisationError("The session is already aligned/cut to the syncregion and can not be cut again")

        if warn_thres is not None:
            sync_start_warn = [
                d.info.sensor_id for d in s.slaves if d._check_sync_packages(warn_thres, where="start") is False
            ]
            sync_end_warn = [
                d.info.sensor_id for d in s.slaves if d._check_sync_packages(warn_thres, where="end") is False
            ]
            if any(sync_end_warn) and cut_end is not True:
                warnings.warn(
                    "For the sensors with the ids {} the last syncpackage occurred more than {} s before the "
                    "end of the dataset. "
                    "The last section of this data should not be trusted.".format(sync_end_warn, warn_thres),
                    SynchronisationWarning,
                )
            if any(sync_start_warn) and cut_start is not True:
                warnings.warn(
                    "For the sensors with the ids {} the first syncpackage occurred more than {} s after the "
                    "start of the dataset. "
                    "The first section of this data should not be trusted.".format(sync_start_warn, warn_thres),
                    SynchronisationWarning,
                )

        # Correct the jump at the beginning of the sync region in the slave counter.
        # This is important because the datasets are later cut based on their first counter value
        for slave in s.slaves:
            if slave.info.sync_index_start <= 1:
                # Unlikely edge case, but let's handle it
                # We do not need to do anything in this case
                continue
            # slave.info.sync_index_start is the first sample (as in # samples) were the index is correct.
            # Therefore, the jump occurs between sample (sync_index_start - 1) and sync_index_start.
            # or equivalently between index (starting at 0) (sync_index_start - 2) and (sync_index_start - 1)
            diff = slave.counter[slave.info.sync_index_start - 1] - slave.counter[slave.info.sync_index_start - 2] - 1
            slave.counter[: slave.info.sync_index_start - 1] += diff

        # Optionally cut to the syncregion
        s = super(SyncedSession, s).cut_to_syncregion(start=cut_start, end=cut_end, inplace=True, warn_thres=None)

        start_idx = [d.counter[0] for d in s.datasets]
        stop_idx = [d.counter[-1] for d in s.datasets]
        if not validate_existing_overlap(np.array(start_idx), np.array(stop_idx)):
            raise ValueError("The provided datasets do not have a overlapping regions!")

        s = super(SyncedSession, s).cut_counter_val(np.max(start_idx), inplace=True)
        stop_idx = [len(d.counter) for d in s.datasets]
        s = super(SyncedSession, s).cut(stop=np.min(stop_idx), inplace=True)

        # Finally set the master counter to all slaves to really ensure identical counters
        for d in s.slaves:
            d.counter = s.master.counter
        s._fully_synced = True
        return s

    def cut_to_syncregion(  # noqa: arguments-differ
        self: Type[T],
        end: bool = False,
        only_to_master: bool = False,
        warn_thres: Optional[int] = 30,
        inplace: bool = False,
    ) -> T:
        """Cut all datasets to the regions where they were synchronised to the master.

        .. deprecated:: 1.0.0
            `cut_to_syncregion` is deprecated and will be replaced by `align_to_syncregion`.
            Use `align_to_syncregion(cut_start=True, ...)` to replicate the functionality of `cut_to_syncregion`.
            `align_to_syncregion` does not support `only_to_master`.

        Parameters
        ----------
        only_to_master :
            If True each slave will be cut to the region, where it was synchronised with the master.
            Master will not be changed. If False, all sensors will be cut to the region, where ALL sensors where
            in sync. Only in the latter case all datasets will have the same length and are guarantied to have the
            same counter.
        end :
            Whether the dataset should be cut at the `info.last_sync_index`. Usually it can be assumed that the
            data will be synchronous for multiple seconds after the last sync package. Therefore, it might be
            acceptable to just ignore the last syncpackage and just cut the start of the dataset.
        warn_thres :
            Threshold in seconds from the end of a dataset. If the last syncpackage occurred more than
            warn_thres before the end of the dataset, a warning is emitted. Use warn_thres = None to silence.
            This is not relevant if the end of the dataset is cut (e.g. `end=True`)
        inplace :
            If operation should be performed on the current Session object, or on a copy
            Info:
        inplace :
            If operation should be performed on the current Session object, or on a copy
            Info:
            `cut_to_syncregion` is deprecated and will be replaced by `align_to_syncregion`.
        inplace :
            If operation should be performed on the current Session object, or on a copy
            Info:
            `cut_to_syncregion` is deprecated and will be replaced by `align_to_syncregion`.
            Use `align_to_syncregion(cut_start=True, ...)` to replicate the functionality of `cut_to_syncregion`.
        inplace :
            If operation should be performed on the current Session object, or on a copy


        """
        warnings.warn(
            '"cut_to_syncregion" is deprecated and will be replaced by "align_to_syncregion". '
            "Use align_to_syncregion(cut_start=True, ...) to replicate the functionality of "
            "cut_to_syncregion.",
            DeprecationWarning,
        )

        if only_to_master is True:
            s = inplace_or_copy(self, inplace)
            s = super(SyncedSession, s).cut_to_syncregion(end=end, inplace=True, warn_thres=None)
            return s

        s = self.align_to_syncregion(cut_start=True, cut_end=end, warn_thres=warn_thres, inplace=inplace)
        return s

    def data_as_df(  # noqa: arguments-differ
        self,
        datastreams: Optional[Sequence[str]] = None,
        index: Optional[str] = None,
        include_units: Optional[bool] = False,
        concat_df: Optional[bool] = False,
    ) -> Union[Tuple["pd.DataFrame"], "pd.DataFrame"]:
        """Export all datasets of the session in a list of (or a single) pandas DataFrame.

        Parameters
        ----------
        datastreams :
            Optional list of datastream names, if only specific ones should be included. Datastreams that
            are not part of the current dataset will be silently ignored.
        index :
            Specify which index should be used for each dataset. The options are:
            "counter": For the actual counter
            "time": For the time in seconds since the first sample
            "utc": For the utc time stamp of each sample
            "utc_datetime": for a pandas DateTime index in UTC time
            None: For a simple index (0...N)
        concat_df :
            If True the individual dfs from each dataset will be concatenated. This is only supported, if the
            session is properly cut to the sync region and all the datasets have the same counter.
        include_units :
            If True the column names will have the unit of the datastream concatenated with an `_`
            Notes:
        include_units :
            If True the column names will have the unit of the datastream concatenated with an `_`
            Notes:
            This method calls the `data_as_df` methods of each Datastream object and then concats the results.
        include_units :
            If True the column names will have the unit of the datastream concatenated with an `_`
            Notes:
            This method calls the `data_as_df` methods of each Datastream object and then concats the results.
            Therefore, it will use the column information of each datastream.

        Returns
        -------
        Session as single or multiple dataframes
            Tuple of pd.DataFrames (one for each Dataset) or a single DataFrame if `concat_df` is set to True

        Raises
        ------
        ValueError
            If any other than the allowed `index` values are used.

        See Also
        --------
        nilspodlib.dataset.Dataset.data_as_df

        """
        import pandas as pd  # noqa: F811

        dfs = super().data_as_df(datastreams, index, include_units=include_units)
        if concat_df is True:
            if not self._fully_synced:
                raise SynchronisationError("Only fully synced datasets, can be exported as a df with unified index.")
            dfs = pd.concat(dfs, axis=1, keys=self.info.sensor_id)
        return dfs

    def imu_data_as_df(  # noqa: arguments-differ
        self, index: Optional[str] = None, include_units: Optional[bool] = False, concat_df: Optional[bool] = False
    ) -> Union[Tuple["pd.DataFrame"], "pd.DataFrame"]:
        """Export the acc and gyro datastreams of all datasets in list of (or a single) pandas DataFrame.

        See Also
        --------
        nilspodlib.session.SyncedSession.data_as_df
        nilspodlib.dataset.Dataset.data_as_df
        nilspodlib.dataset.Dataset.imu_data_as_df

        Parameters
        ----------
        index :
            Specify which index should be used for the dataset. The options are:
            "counter": For the actual counter
            "time": For the time in seconds since the first sample
            "utc": For the utc time stamp of each sample
            "utc_datetime": for a pandas DateTime index in UTC time
            None: For a simple index (0...N)
        concat_df :
            If True the individual dfs from each dataset will be concatenated. This is only supported, if the
            session is properly cut to the sync region and all the datasets have the same counter.
        include_units :
            If True the column names will have the unit of the datastream concatenated with an `_`
            Notes:
        include_units :
            If True the column names will have the unit of the datastream concatenated with an `_`
            Notes:
            This method calls the `data_as_df` methods of each Datastream object and then concats the results.
        include_units :
            If True the column names will have the unit of the datastream concatenated with an `_`
            Notes:
            This method calls the `data_as_df` methods of each Datastream object and then concats the results.
            Therefore, it will use the column information of each datastream.

        Returns
        -------
        Imu data as single or multiple dataframes
            Tuple of pd.DataFrames (one for each Dataset) or a single DataFrame if `concat_df` is set to True

        Raises
        ------
        ValueError
            If any other than the allowed `index` values are used.

        """
        return self.data_as_df(
            datastreams=["acc", "gyro"], index=index, include_units=include_units, concat_df=concat_df
        )



# -*- coding: utf-8 -*-
"""Header class(es), which are used to read and access basic information from a recorded session."""

import datetime
import json
import pprint
import warnings
from itertools import chain
from collections import OrderedDict
from distutils.version import StrictVersion
from typing import Tuple, Any, List, Dict, Union

import numpy as np

from nilspodlib.consts import SENSOR_POS
from nilspodlib.utils import convert_little_endian


class _HeaderFields:
    """Base class listing all the attributes of a session header.

    For documentation see the `Header` object.
    """

    enabled_sensors: Tuple[str]

    motion_interrupt_enabled: bool
    dock_mode_enabled: bool
    sensor_position: str
    session_termination: str
    sample_size: int

    sampling_rate_hz: float
    acc_range_g: float
    gyro_range_dps: float

    sync_role: str
    sync_distance_ms: float
    sync_address: int
    sync_channel: int
    sync_index_start: int
    sync_index_stop: int

    utc_start: int
    utc_stop: int

    version_firmware: str
    version_hardware: str
    mac_address: str

    custom_meta_data: Tuple[float]

    n_samples: int

    # Note this must correspond to the order they appear in the datapackage when activated
    _SENSOR_FLAGS = OrderedDict(
        [
            ("gyro", (0x02, 0x00)),
            ("acc", (0x01, 0x00)),
            ("mag", (0x04, 0x00)),
            ("baro", (0x08, 0x00)),
            ("analog", (0x10, 0x00)),
            ("ecg", (0x20, 0x00)),
            ("ppg", (0x40, 0x00)),
            ("temperature", (0x80, 0x00)),
        ]
    )

    _OPERATION_MODES = {
        "motion_interrupt_enabled": 0x80,
        "dock_mode_enabled": 0x40,
    }

    _SAMPLING_RATES = {1: 1024.0, 2: 512.0, 4: 256.0, 5: 204.8, 10: 102.4, 20: 51.2, 40: 25.6, 80: 12.8, 160: 6.4}

    _SESSION_TERMINATION = {"no memory": 0x10, "BLE": 0x20, "dock": 0x40, "low battery": 0x80}

    _SYNC_ROLE = {0: "disabled", 1: "slave", 2: "master"}

    @property
    def _header_fields(self) -> List[str]:
        """List all header fields.

        This is a little hacky and relies on that the header fields are the only attributes that are type annotated.
        """
        return list(_HeaderFields.__annotations__.keys())

    @property
    def _all_header_fields(self) -> List[str]:
        """ """
        additional_fields = [
            "duration_s",
            "utc_datetime_start",
            "utc_datetime_stop",
            "utc_datetime_start_day_midnight",
            "is_synchronised",
            "has_position_info",
            "sensor_id",
            "strict_version_firmware",
        ]

        return self._header_fields + additional_fields

    @property
    def duration_s(self) -> int:
        """Length of the measurement."""
        return self.utc_stop - self.utc_start

    @property
    def utc_datetime_start(self) -> datetime.datetime:
        """Start time as utc datetime."""
        return datetime.datetime.utcfromtimestamp(self.utc_start)

    @property
    def utc_datetime_stop(self) -> datetime.datetime:
        """Stop time as utc datetime."""
        return datetime.datetime.utcfromtimestamp(self.utc_stop)

    @property
    def utc_datetime_start_day_midnight(self) -> datetime.datetime:
        """UTC timestamp marking midnight of the recording date."""
        return datetime.datetime.combine(self.utc_datetime_start.date(), datetime.time(), tzinfo=datetime.timezone.utc)

    @property
    def is_synchronised(self) -> bool:
        """If a recording was synchronised or not."""
        return not self.sync_role == "disabled"

    @property
    def has_position_info(self) -> bool:
        """If any information about the sensor position is provided."""
        return not self.sensor_position == "undefined"

    @property
    def sensor_id(self) -> str:
        """Get the unique sensor identifier."""
        return "".join(self.mac_address[-5:].split(":"))

    @property
    def strict_version_firmware(self) -> StrictVersion:
        """Get the firmware as a StrictVersion object."""
        return StrictVersion(self.version_firmware[1:])

    def __str__(self) -> str:
        full_header = {k: getattr(self, k, None) for k in self._all_header_fields}
        return pprint.pformat(full_header)


class Header(_HeaderFields):
    """Additional Infos of recording.

    Usually their is no need to use this class on its own, but it is just used as a convenient wrapper to access all
    information via a dataset instance.

    .. warning :: UTC timestamps and datetime, might not be in UTC. We just provide the values recorded by the sensor
                  without any local conversions.
                  Depending on the recording device, a localized timestamp might have been set for the internal sensor
                  clock

    Attributes
    ----------
    sensor_id
        Get the unique sensor identifier.
    enabled_sensors
        Tuple of sensors that were enabled during the recording.
        Uses typical shorthands.
    motion_interrupt_enabled
        If the motion interrupt of the sensor was enable.
    dock_mode_enabled
        If the sensor was used in dock/home monitoring mode (charging starts/stops sessions)
    sensor_position
        If a sensor position was specified.
        Can be a position from a list or custom bytes.
    has_position_info
        If any information about the sensor position is provided.
    session_termination
        The reason, why the session was stopped.
        Can be used to diagnose potential issues.
    sample_size
        The size of a single sample in bytes.
        This depends on the enabled sensors.
    sampling_rate_hz
        Sampling rate of the recording.
    acc_range_g
        Range of the accelerometer in g.
    gyro_range_dps
        Range of the gyroscope in deg per s.
    sync_role
        Can be "master", "slave" or disabled.
    is_synchronised
        If a recording was synchronised or not.

        .. note:: This does only indicate, that the session was recorded with the sync feature enabled, not that the
                  data is actually synchronised.
    sync_distance_ms
        How far apart the sync packages were set to be.
        This is only a lower limit and does not represent the actual timing.
    sync_address
        Shared address used by all sensors of the sync group.
    sync_channel
        Shared radio channel used by all sensors of the sync group.
    sync_index_start
        Index (position in counter) were the first sync package was received (0 if sync_role=="master")
    sync_index_stop
        Index (position in counter) were the last sync package was received (0 if sync_role=="master")
    utc_start
        Unix time stamp of the start of the recording.

        .. note:: No timezone is assumed and client software is instructed to set the internal sensor clock to utc time.
                  However, this can not be guaranteed.
    utc_datetime_start
        Start time as utc datetime.
    utc_datetime_stop
        Stop time as utc datetime.
    utc_stop
        Unix time stamp of the end of the recording.

        .. note:: No timezone is assumed and client software is instructed to set the internal sensor clock to utc time.
                  However, this can not be guaranteed.
    utc_datetime_start_day_midnight
        UTC timestamp marking midnight of the recording date.
        This is useful, as the sensor internal counter gets reset at midnight.
        I.e. utc_datetime_start_day_midnight + counter[0] * sampling_rate should be utc_datetime_start
    duration
        Length of the measurement in seconds.
    version_firmware
        Version number of the firmware
    strict_version_firmware
        The firmware as a StrictVersion object.
    version_hardware
        Hardware revision of the sensor.
        Another revision usually indicates, that data should be handled differently.
    mac_address
        BLE Mac address of the sensor.
    custom_meta_data
        Custom meta data (3 bytes) which was saved on the sensor.
    n_samples
        Number of samples recorded during the measurement

        .. note:: The number of samples might not be equal to the actual number of samples in the file, because the
                  sensor only transmits full flash pages.
                  This means a couple of samples (max. 2048/sample_size) at the end might be cut.
                  This issue should be resolved in firmware versions 0.14.0 or higher.

    """

    def __init__(self, **kwargs):
        """Initialize a header object.

        This will just put all values provided in kwargs as attributes onto the class instance.
        If one value has an unexpected name, a warning is raised, and the key is ignored.
        """
        for k, v in kwargs.items():
            if k in self._header_fields:
                setattr(self, k, v)
            else:
                # Should this be a error?
                warnings.warn("Unexpected Argument {} for Header".format(k))

    @classmethod
    def from_bin_array(cls, bin_array: np.ndarray) -> "Header":
        """Create a new Header instance from an array of bytes."""
        header_dict = cls.parse_header_package(bin_array)
        return cls(**header_dict)

    @classmethod
    def from_json(cls, json_string: str) -> "Header":
        """Create a new Header from a json export of the header.

        This is only tested with the direct output of the `to_json` method and should only be used to reimport a Header
        exported with this method.
        """
        h = cls(**json.loads(json_string))
        # ensure that the enabled sensors and custom_metadata have the right dtype
        for k in ("enabled_sensors", "custom_meta_data"):
            if getattr(h, k, None):
                setattr(h, k, tuple(getattr(h, k)))
        return h

    def to_json(self) -> str:
        """Export a header as json.

        It can be imported again using the `from_json` method without information loss.
        """
        header_dict = {k: v for k, v in self.__dict__.items() if k in self._header_fields}
        return json.dumps(header_dict)

    @classmethod
    def parse_header_package(cls, bin_array: np.ndarray) -> Dict[str, Union[str, int, float, bool, tuple]]:
        """Extract all values from a binary header package."""
        # Note that because the info packet already has the first byte (info size) removed, all byte numbers are
        # shifted compared to the documentation
        bin_array = bin_array.astype(np.uint32)
        header_dict = dict()

        header_dict["sample_size"] = int(bin_array[0])

        sensors = bin_array[1:3]
        enabled_sensors = list()
        for para, val in cls._SENSOR_FLAGS.items():
            if bool(sensors[0] & val[0]) or bool(sensors[1] & val[1]):
                enabled_sensors.append(para)
        header_dict["enabled_sensors"] = tuple(enabled_sensors)

        # bin_array[2] = currently not used

        header_dict["sampling_rate_hz"] = cls._SAMPLING_RATES[bin_array[3]]

        header_dict["session_termination"] = next(
            k for k, v in cls._SESSION_TERMINATION.items() if bool(bin_array[4] & v) is True
        )

        header_dict["sync_role"] = cls._SYNC_ROLE[bin_array[5]]

        header_dict["sync_distance_ms"] = float(bin_array[6] * 100.0)

        header_dict["acc_range_g"] = float(bin_array[7])

        header_dict["gyro_range_dps"] = float(bin_array[8] * 125)

        sensor_position = bin_array[9]
        header_dict["sensor_position"] = SENSOR_POS.get(sensor_position, sensor_position)

        operation_mode = bin_array[10]
        for para, val in cls._OPERATION_MODES.items():
            header_dict[para] = bool(operation_mode & val)

        header_dict["custom_meta_data"] = tuple(bin_array[11:14].astype(float))

        # Note: We ignore timezones and provide just the time info, which was stored in the sensor
        header_dict["utc_start"] = int(convert_little_endian(bin_array[14:18]))
        header_dict["utc_stop"] = int(convert_little_endian(bin_array[18:22]))

        header_dict["n_samples"] = int(convert_little_endian(bin_array[22:26]))

        header_dict["sync_index_start"] = int(convert_little_endian(bin_array[26:30]))
        header_dict["sync_index_stop"] = int(convert_little_endian(bin_array[30:34]))

        header_dict["mac_address"] = ":".join(["{:02x}".format(int(x)) for x in bin_array[34:40]][::-1])

        header_dict["sync_address"] = "".join(["{:02x}".format(int(x)) for x in bin_array[40:45]][::-1])
        header_dict["sync_channel"] = int(bin_array[45])

        header_dict["version_hardware"] = "".join((str(x) for x in bin_array[46:48]))

        header_dict["version_firmware"] = "v{}.{}.{}".format(*(int(x) for x in bin_array[-3:]))

        return header_dict


class _ProxyHeader(_HeaderFields):
    """A proxy header used by session objects to get direct access to multiple headers.

    This allows to access attributes of multiple header instances without reimplementing all of its attributes.
    This is achieved by basically intercepting all getattribute calls and redirecting them to all header instances.

    This concept only allows read only access. However, usually their is no need to modify a header after creation.
    """

    _headers: Tuple[Header]

    def __init__(self, headers: Tuple[Header]):
        self._headers = headers

    def __getattribute__(self, name: str) -> Tuple[Any]:
        if name in ("_headers", "_all_header_fields", "_header_fields", "_ipython_display_"):
            return super().__getattribute__(name)
        if callable(getattr(self._headers[0], name)) is True:
            if name.startswith("__"):
                return super().__getattribute__(name)
            raise ValueError(
                "_ProxyHeader only allows access to attributes of the info objects. {} is a callable method.".format(
                    name
                )
            )

        return tuple([getattr(d, name) for d in self._headers])

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "_headers":
            return super().__setattr__(name, value)
        raise NotImplementedError("_ProxyHeader only allows readonly access to the info objects of a dataset")

    def __dir__(self):
        return chain(super().__dir__(), self._headers[0].__dir__())

    def _ipython_display_(self):
        """ """
        import pandas as pd  # noqa: import-outside-toplevel
        from IPython import display  # noqa: import-outside-toplevel

        header = {k: getattr(self, k, None) for k in self._all_header_fields}
        display.display(pd.DataFrame(header, index=self.sensor_id).T)



"""Set of helper functions used throughout the library."""

import copy
import struct
import warnings
from distutils.version import StrictVersion
from pathlib import Path
from typing import TypeVar, Tuple, Any

import numpy as np

from nilspodlib.exceptions import CorruptedPackageWarning

path_t = TypeVar("path_t", str, Path)  # noqa: invalid-name
T = TypeVar("T")


def convert_little_endian(byte_list: np.ndarray, dtype: Any = int) -> np.ndarray:
    """Convert a little endian bytestring into a readable format.

    Parameters
    ----------
    byte_list :
        The array of bytes
    dtype :
        The datatype, the final value should be converted to.

    """
    byte_list = np.array(byte_list).astype(np.uint32)
    number = byte_list[0]
    for i, v in enumerate(byte_list[1:]):
        number |= v << int(8 * (i + 1))
    return number.astype(dtype)


def read_binary_uint8(data_bytes: np.ndarray, packet_size: int, expected_samples: int) -> np.ndarray:
    """Read a continuous stream of uint8 values into its separate datapoints.

    Parameters
    ----------
    data_bytes :
        The raw stream of data bytes
    packet_size :
        The size of each datapacket stored in the stream
    expected_samples :
        The expected number of samples in the data stream. This is only used to check the integrity
        of the datastream and raise a warning if, the number of samples in the dataset, does not match the expected
        number.

    Returns
    -------
    data
        Array with the shape (n_samples, packet_size)

    """
    packet_size = int(packet_size)
    expected_length = expected_samples * packet_size
    page_size = 2048
    if abs(len(data_bytes) - expected_length) > page_size // packet_size:
        warnings.warn(
            "The provided binary file contains more or less than {0} packages than indicated by the header"
            " ({1} vs. {2}). This can be caused by a bug affecting all synchronised sessions recorded with"
            " firmware versions before 0.14.0. \n"
            "The full file will be read to avoid data loss, but this might add up to {0} corrupted packages"
            " at the end of the datastream.".format(
                page_size // packet_size, expected_samples, len(data_bytes) // packet_size
            ),
            CorruptedPackageWarning,
        )
        expected_length = (len(data_bytes) // packet_size) * packet_size

    elif expected_length > len(data_bytes):
        warnings.warn(
            "The provided binary file contains less samples than indicated by the header."
            " This might mean that the file was corrupted.",
            CorruptedPackageWarning,
        )
        expected_length = (len(data_bytes) // packet_size) * packet_size

    data_bytes = data_bytes[:expected_length]
    data = np.reshape(data_bytes, (expected_length // packet_size, packet_size))
    return data


def get_header_and_data_bytes(path: path_t) -> Tuple[np.ndarray, np.ndarray]:
    """Separate a binary file into its header and data part."""
    with open(path, "rb") as f:
        header = f.read(1)
        header_size = header[0]
        header += f.read(header_size - 1)
        data_bytes = np.fromfile(f, dtype=np.dtype("B"))

    header = bytearray(header)
    header_bytes = np.asarray(struct.unpack(str(header_size) + "b", header[0:header_size]), dtype=np.uint8)

    return header_bytes, data_bytes


def get_sample_size_from_header_bytes(header_bytes: np.ndarray) -> int:
    """Get the size of an individual data sample (in bytes) from the header info."""
    return int(header_bytes[1])


def get_strict_version_from_header_bytes(header_bytes: np.ndarray) -> StrictVersion:
    """Extract the version number from a byte header."""
    return StrictVersion("{}.{}.{}".format(*(int(x) for x in header_bytes[-3:])))


def inplace_or_copy(obj: T, inplace: bool) -> T:
    """Either create a deepcopy of the object or return it based on the value of `inplace`."""
    if inplace is True:
        return obj
    return copy.deepcopy(obj)


def validate_existing_overlap(start_vals: np.ndarray, end_vals: np.ndarray) -> bool:
    """Check that multiple intervals indicated by their start and stop values do all have an overlapping region.

    Raises
    ------
    ValueError
        If any of the intervals are invalid, because their end values is before their start value.

    """
    if not all(i < j for i, j in zip(start_vals, end_vals)):
        raise ValueError("The start values need to be smaller then their respective end values!")
    return np.max(start_vals) < np.min(end_vals)


def remove_docstring_indent(doc_str: str) -> str:
    """Remove the additional indent of a multiline docstring.

    This can be helpful, if docstrings are combined programmatically.
    """
    lines = doc_str.split("\n")
    if len(lines) <= 1:
        return doc_str
    first_non_summary_line = next(line for line in lines[1:] if line)
    indent = len(first_non_summary_line) - len(first_non_summary_line.lstrip())
    cut_lines = [lines[0]]
    for line in lines:
        cut_lines.append(line[indent:])
    return "\n".join(cut_lines)



"""Exceptions and warnings used in the library."""
import warnings


class InvalidInputFileError(Exception):
    """Indicate an invalid binary file, which can not be loaded."""


class RepeatedCalibrationError(Exception):
    """Indicate that a datastream was already calibrated."""

    MESSAGE = 'The sensor "{}" is already calibrated. Repeated calibration will lead to wrong values.'

    def __init__(self, sensor_name):
        """Get a new Exception instance."""
        message = self.MESSAGE.format(sensor_name)
        super().__init__(message)


class SynchronisationError(Exception):
    """Error that is raised for sync related issues."""


class SynchronisationWarning(Warning):
    """Indicate potential issues with sync."""


class LegacyWarning(Warning):
    """Indicate potential issues due to older firmware versions."""


class CorruptedPackageWarning(Warning):
    """Indicate potential issues with a recorded file."""


class CalibrationWarning(Warning):
    """Indicate potential issues with a calibration."""


class VersionError(Exception):
    """Error related to Firmware Version issues."""


def datastream_does_not_exist_warning(sensor_type, operation):
    """Warn about not existing datastreams."""
    message = 'The datastream "{}" does not exist for the current session.\
     The performed operation "{}" will have not effect'.format(
        sensor_type, operation
    )
    return warnings.warn(message)
